
import json
import nvgpu
import docker
import logging
import platform
import requests
import subprocess
from os.path import expanduser, join, exists

OAUTH_API_URL = 'https://api.nanami.io/oauth'
# QUEUE_API_URL = 'https://api.nanami.io/queue'
QUEUE_API_URL = 'http://localhost:5000'

logger = logging.getLogger(__name__)
logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
    level=logging.INFO,
)

def container_stop(containers):
    for container in containers:
        if container:
            container.stop()
            logger.info(f"{container}: is stopped.")

def container_run(client, arguments, command, setup_command=None):
    container = client.containers.create(**arguments)

    logger.info(f'Image : {arguments["image"]} is now starting in {container}!')

    container.start()
    try:
        nanami_setup_log = container.exec_run('pip install -U nanami', stream=True, tty=True, environment=["PYTHONUNBUFFERED=0"])
        for line in nanami_setup_log[1]:
            logger.info(line.decode('utf-8').strip())

        if setup_command:
            setup_log = container.exec_run(setup_command, stream=True, tty=True, environment=["PYTHONUNBUFFERED=0"])
            for line in setup_log[1]:
                logger.info(line.decode('utf-8').strip())

        exec_log = container.exec_run(command, stream=True, tty=True, environment=["PYTHONUNBUFFERED=0"])
        for line in exec_log[1]:
            logger.info(line.decode('utf-8').strip())

    except KeyboardInterrupt as e:
        logger.info(f'Container will be stopped!')
        container_stop([container])

def check_arguments(data, arguments):
    errors = []
    for _argument in arguments:
        if _argument not in data:
            errors.append(f"Argument({_argument}) is not found.")
    if errors:
        raise NotImplementedError(errors)

def check_auth():
    path = join(expanduser("~"), '.nanami.json')
    if not exists(path):
        raise FileNotFoundError("Please `nanami login` first.")
    with open(path) as fp:
        user = json.load(fp)
    return user

def check_entrypoint(entrypoint):
    with open(entrypoint) as f:
        start, end = False, False
        for line in f.read().splitlines():
            line = line.strip()
            if "nanami.start()" == line:
                start = True
            if "nanami.end()" == line:
                end = True
    return (start and end)

def docker_version_check():
    client = docker.from_env()
    version = [int(v) for v in client.version()['Components'][0]['Version'].split('.')]
    assert version[0] >= 19 and version[1] >= 3

def find_image_in_local(image):
    client = docker.from_env()
    images = []
    for _image in client.images.list():
        images.extend(_image.tags)
    if image in images:
        return True
    else:
        return False

def gpu_check(gpus='', verbose=False):
    if gpus == '':
        return None

    device_requests = []
    nvidia_smi = parse_nvidia_smi()
    if nvidia_smi:
        cuda_version = nvidia_smi['CUDA Version']
        if gpus == 'all':
            gpus = nvgpu.available_gpus()
            device_requests = [
                docker.types.DeviceRequest(count=-1, capabilities=[['gpu']])
            ]
        else:
            device_requests = []
            gpu_uuid = dict({g['index'] : g['uuid'] for g in nvgpu.gpu_info()})
            for g_idx in gpus.split(','):
                if g_idx not in [g['index'] for g in nvgpu.gpu_info()]:
                    raise NotImplementedError(f"Device id({g_idx}) is avaliable.")
                elif g_idx not in nvgpu.available_gpus():
                    raise NotImplementedError(f"Device id({g_idx}) is using.")
                device_requests.extend([
                    docker.types.DeviceRequest(
                        driver='nvidia',
                        device_ids=[gpu_uuid[g_idx]],
                        capabilities=[
                            ['gpu']
                        ]
                    )
                ])
        if gpus:
            client = docker.from_env()
            nvidia_smi = client.containers.run(
                f'nvidia/cuda:{cuda_version}-base',
                'nvidia-smi',
                remove=True,
                device_requests=device_requests
            )
            if verbose:
                print(nvidia_smi.decode('utf-8'))
        else:
            device_requests = []
            logger.info('All GPUs is using.')

    return device_requests

def image_search(image):
    sp = subprocess.Popen(
        ['docker', 'inspect', '--type=image', image], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out_str = sp.communicate()
    out_json = json.loads(out_str[0].decode("utf-8"))
    if out_json:
        return True
    else:
        return False

def image_pulling(image):
    completed_process = subprocess.run(
        ['docker', 'pull', image], input=None, stdout=None, stderr=None
    )
    if completed_process.returncode != 0:
        raise NotImplementedError(f"Pulling Error : {image}")

def nvidia_docker_version_check():
    try:
        sp = subprocess.Popen(['nvidia-docker', 'version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        out_str = sp.communicate()
        out_list = out_str[0].decode("utf-8").split('\n')

        out_dict = {}

        for item in out_list:
            try:
                key, val = item.split(':')
                key, val = key.strip(), val.strip()
                out_dict[key] = val
            except:
                pass

        version = [int(v) for v in out_dict['NVIDIA Docker'].split('.')]
        assert version[0] >= 2 and version[1] >= 5
    except:
        logger.warning("nvidia-docker is not installed.")

def parse_nvidia_smi():
    try:
        sp = subprocess.Popen(['nvidia-smi', '-q'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except:
        return None

    out_str = sp.communicate()
    out_list = out_str[0].decode("utf-8").split('\n')

    out_dict = {}

    for item in out_list:
        try:
            key, val = item.split(':')
            key, val = key.strip(), val.strip()
            out_dict[key] = val
        except:
            pass

    return out_dict

def post(path, data):
    r = requests.post(path, json=data)
    if r.status_code == 200:
        return r.json()
    else:
        return None



import sys
from pathlib import Path

from nanami.utils import *

class Nanami(object):
    def __init__(self, image, workspace, entrypoint, gpus='', argument='', cache=False, queue_id=None):
        self.image = image

        self.workspace = Path(workspace).absolute()
        self._check_workspace()

        self.container_entrypoint = join('/workspace', entrypoint)
        self.entrypoint = join(self.workspace, entrypoint)
        self._check_entrypoint()

        self.requirements = join(self.workspace, 'requirements.txt')
        self._check_requirements()
        self.argument = argument
        self.command = f"python {self.container_entrypoint} {argument}"
        self.setup_command = f"pip install -r {join('/workspace', 'requirements.txt')}"

        self.gpus = gpus
        self.device_requests = gpu_check(gpus=gpus, verbose=False)
        self.cache = cache
        self.queue_id = queue_id
        self._check_image()

    def _check_entrypoint(self):
        if not exists(self.entrypoint):
            raise FileNotFoundError(self.entrypoint)
        if not check_entrypoint(self.entrypoint):
            raise NotImplementedError(f"Can't find nanami.start() and nanami.end() in {self.entrypoint}.")

    def _check_requirements(self):
        _requirements = join(self.workspace, 'requirements.txt')
        if exists(_requirements):
            with open(_requirements) as f:
                _requirements_list = f.read().splitlines()
            logger.info(f'Found requirements.txt: {_requirements_list}')
        else:
            logger.info('Not found requirements.txt')

    def _check_workspace(self):
        if not exists(self.workspace):
            raise FileNotFoundError(self.workspace)

    def _check_image(self):
        if not image_search(image=self.image):
            raise NotImplementedError(f"{self.image} is not found.")

    def save(self, apikey, port):
        if sys.platform == 'linux':
            localhost = f"http://127.0.0.1:{port}"
        elif sys.platform == "darwin":
            localhost = f'http://host.docker.internal:{port}'

        with open(join(self.workspace, 'nanami.json'), 'w') as fp:
            json.dump({
                'apikey': apikey,
                'queue_info' : {
                    'queue_id': self.queue_id,
                    'image': self.image,
                    'workspace': str(self.workspace),
                    'entrypoint': str(self.entrypoint),
                    'gpus': self.gpus,
                    'argument': self.argument,
                    'cache': self.cache,
                },
                'localhost': localhost
            }, fp, indent=4)


import os
from setuptools import setup, find_packages

project_name = "nanami"
version = os.environ.get('COMMIT_VERSION', '0.0.0')

with open('requirements.txt') as f:
    required = f.read().splitlines()

if __name__ == "__main__":

    with open('README.md', 'r') as t:
        README = t.read()

    setup(
        name=project_name,
        version=version,
        long_description=README,
        long_description_content_type='text/markdown',
        license="",
        author="nanami-ci",
        author_email="gray@nanami.io",
        description="",
        url="https://github.com/nanami-ci/client",
        python_requires=">=3.6",
        platforms=["any"],
        project_urls={
            "Source Code": "https://github.com/nanami-ci/client",
        },
        install_requires=required,
        entry_points={
            'console_scripts': [
                'nanami=nanami.main:cli'
            ],
        },
        packages=find_packages(),
    )



import sys
import json
import atexit
import logging
from io import StringIO

from nanami.utils import logger, post, QUEUE_API_URL, check_auth

log_stream = StringIO()
error_logger = logging.getLogger()
logging.basicConfig(stream=log_stream)

def at_exit_func():
    error_message = log_stream.getvalue().strip()
    if len(error_message) > 0:
        project = set_status(status='failure', error_message=error_message)
        if project:
            post(f"{project['localhost']}/message", data={
                'project': project,
                'status': 'failure',
            })
            post(f"{project['localhost']}/running", data={
                'project': project,
            })


def exception_hook(exc_type, exc_value, exc_traceback):
    error_logger.error(
        "Uncaught exception",
        exc_info=(exc_type, exc_value, exc_traceback)
    )

sys.excepthook = exception_hook
atexit.register(at_exit_func)

def read_project_info():
    try:
        with open('/workspace/nanami.json') as json_file:
            project_info = json.loads(json_file.read())
            return project_info
    except:
        raise FileNotFoundError("nanami.json is not found.")

def set_status(status, error_message=None):
    project = read_project_info()
    apikey = project['apikey']

    result = post(f"{QUEUE_API_URL}/status", data={
        'apikey': apikey,
        'queue_info': project['queue_info'],
        'status': status,
        'error_message': error_message
    })
    if result:
        if not result['success']:
            print(f'Failed to set {status}.')
        else:
            if result['dequeue']:
                logger.info("Nanami will dequeue queue_id.")
                return project
    else:
        print(f'[API Error] : {QUEUE_API_URL}/status')

    return None

def start():
    logger.info('[Nanami Start!]')
    project = set_status(status='running')
    if project:
        post(f"{project['localhost']}/message", data={
            'project': project,
            'status': 'running',
        })

def end():
    logger.info('[Nanami End!]')
    project = set_status(status='success')
    if project:
        post(f"{project['localhost']}/message", data={
            'project': project,
            'status': 'success',
        })
        post(f"{project['localhost']}/running", data={
            'project': project,
        })



import json
import click
from flask import Flask, request
from urllib.parse import urlparse

from nanami.utils import *
from nanami.nanami import Nanami

proxy_container = None
app = Flask(__name__)

docker_version_check()
nvidia_docker_version_check()

def create_app():
    @app.route('/')
    def index():
        return 'Nanami Test'

    @app.route('/message', methods=['POST'])
    def message():
        data = request.get_json()
        check_arguments(data, ["project", "status"])
        logger.info(f'[{data["status"]}]: {data["project"]}')

    @app.route('/running', methods=['POST'])
    def running():
        data = request.get_json()
        check_arguments(data, ["project"])

        port = urlparse(data['project']['localhost']).port
        dequeuing(QUEUE_API_URL, apikey=data['project']['apikey'], port=port)

    return app

def dequeuing(QUEUE_API_URL, apikey, port):
    dequeue_result = post(f"{QUEUE_API_URL}/dequeue", data={'apikey': apikey})
    if dequeue_result:
        if dequeue_result['success']:
            queue_info = dequeue_result['queue_info']
            running(
                image=queue_info['image'],
                workspace=queue_info['workspace'],
                entrypoint=queue_info['entrypoint'],
                gpus=queue_info['gpus'],
                argument=queue_info['argument'],
                cache=queue_info['cache'],
                queue_id=queue_info['queue_id'],
                apikey=apikey,
                port=port
            )
        else:
            if dequeue_result['message']:
                print(f'Failed to dequeue : {(dequeue_result["message"])}.')
            else:
                print(f'Failed to dequeue.')
    else:
        print(f'[API Error] : {QUEUE_API_URL}/dequeue')

def running(image, workspace, entrypoint, gpus, argument, cache, queue_id=None, apikey=None, port=None):
    nanami = Nanami(image, workspace, entrypoint, gpus, argument, cache, queue_id)
    if queue_id and apikey and port:
        nanami.save(apikey=apikey, port=port)

    if not find_image_in_local(image=image):
        image_pulling(image=image)

    run_argument = {
        'image': image,
        'command': "/bin/bash",
        'volumes': {nanami.workspace: {'bind': '/workspace', 'mode': 'rw'}},
        'working_dir': '/workspace',
        'tty': True,
        'stdin_open': True,
    }

    client = docker.from_env()
    if nanami.device_requests:
        logger.info('[GPU mode]')
        run_argument.update({'device_requests': nanami.device_requests})
        container_run(client=client, arguments=run_argument, command=nanami.command, setup_command=nanami.setup_command)
    else:
        if gpus:
            logger.warning("You are running with GPU mode but All GPUs is working.")
        logger.info('[CPU mode]')
        container_run(client=client, arguments=run_argument, command=nanami.command, setup_command=nanami.setup_command)

@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    if not ctx.invoked_subcommand:
        pass

@cli.command()
@click.option('--gpus', '-g', type=str, default='all', help='GPU number ex) 0,1')
def gpu(gpus):
    gpu_check(gpus=gpus, verbose=True)

@cli.command()
@click.option('--image', '-i', type=str, required=True, help='Docker image name')
@click.option('--workspace', '-w', type=str, required=True, help='Workspace directory path')
@click.option('--entrypoint', '-e', type=str, required=True, help='Entrypoint filename')
@click.option('--gpus', '-g', type=str, default='', help='GPU number ex) 0,1')
@click.option('--argument', '-a', type=str, default='')
@click.option('--cache', '-c', is_flag=True, help="Caching Docker image")
# TODO: caching mechanism with docker image build
def local(image, workspace, entrypoint, gpus='', argument='', cache=False):
    running(image, workspace, entrypoint, gpu, argument, cache)

@cli.command()
def queue():
    user = check_auth()
    queue_list = post(f"{QUEUE_API_URL}/queue", data={
        'apikey': user['user']['apikey'],
    })
    if queue_list:
        if queue_list['success']:
            print(f'enqueue succeeded : {queue_list["queue_list"]}')
        else:
            print('enqueue failed.')
    else:
        print(f'[API Error] : {QUEUE_API_URL}/queue.')

@cli.command()
@click.option('--image', '-i', type=str, required=True, help='Docker image name')
@click.option('--workspace', '-w', type=str, required=True, help='Workspace directory path')
@click.option('--entrypoint', '-e', type=str, required=True, help='Entrypoint filename')
@click.option('--gpus', '-g', type=str, default='', help='GPU number ex) 0,1')
@click.option('--argument', '-a', type=str, default='')
@click.option('--cache', '-c', is_flag=True, help="Caching Docker image")
def add(image, workspace, entrypoint, gpus='', argument='', cache=False):
    user = check_auth()
    Nanami(image, workspace, entrypoint, gpus, argument, cache)
    enqueue_result = post(f"{QUEUE_API_URL}/add", data={
        'apikey' : user['user']['apikey'],
        'image': image,
        'workspace': workspace,
        'entrypoint': entrypoint,
        'gpus': gpus,
        'argument': argument,
        'cache': cache
    })
    if enqueue_result:
        if enqueue_result['success']:
            print(f'enqueue succeeded : {enqueue_result["queue_id"]}.')
        else:
            print('enqueue failed.')
    else:
        print(f'[API Error] : {QUEUE_API_URL}/add.')

@cli.command()
@click.option('--id', type=str, required=True, help='Queue id')
def rm(id):
    user = check_auth()
    remove_result = post(f"{QUEUE_API_URL}/remove", data={
        'apikey': user['user']['apikey'],
        'queue_id': id
    })
    if remove_result:
        if remove_result['success']:
            print(f'Successful removal of queue_id({id}).')
        else:
            print(f'Failed to remove queue_id({id}).')
    else:
        print(f'[API Error] : {QUEUE_API_URL}/remove.')

@cli.command()
@click.option('--port', type=int, default=7007, help='')
def start(port):
    user = check_auth()
    path = join(expanduser("~"), '.nanami.json')
    with open(path, 'w') as fp:
        json.dump({'user': user['user'], 'port': port}, fp, indent=4)
    apikey = user['user']['apikey']

    dequeuing(QUEUE_API_URL, apikey, port)

    app = create_app()
    app.run('0.0.0.0', port=port)

@cli.command()
@click.option('--apikey', type=str, default=None,
    help='You can find your API key in your browser here: https://nanami.io/')
def login(apikey):
    if apikey:
        user = post(f"{OAUTH_API_URL}/check", data={'apikey': apikey})
        if user:
            if user['success']:
                user = json.loads(user['user'])
                print(f"Welcome {user['github_login']}!")
                path = join(expanduser("~"), '.nanami.json')
                with open(path, 'w') as fp:
                    json.dump({'user': user}, fp, indent=4)
            else:
                print('Authentication is fail.')
        else:
            print(f'[API Error] : {OAUTH_API_URL}/check')
    else:
        print('`apikey` should not `None`. "Try nanami login --help" for help.')